/*
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "Ant" and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package org.apache.tools.ant.taskdefs;

import sun.misc.UUEncoder;

import java.io.File;
import java.io.FileOutputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Vector;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

import org.apache.tools.ant.types.FileSet;
import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.BuildException;

/**
 * Creates a shar file.  A shar file is a UNIX shell script that extracts one or
 * more files from within its body and optionally executes additional shell
 * commands.
 * @since Ant1.6
 * @ant.task name="shar" category="packaging"
 * @author Magesh Umasankar
 */
public final class Shar extends Task {
    /** The shar file */
    private File destFile;
    /**
     * Additional shell commands that are to be executed after the
     * self-extraction is completed.
     */
    protected Postscript postscript;
    /** Files that are to be archived */
    protected Vector filesets = new Vector();
    /** UUEncode the files that are being archived?  Defaults to true */
    private boolean uuencode = true;
    private static final char[] hexChars =
            {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};
    private static final String HEADER =
            "# This is a shell archive.  Remove anything before\n"
            + "# this line, and then unpack it by entering \"sh file\".  Note, it may\n"
            + "# create directories; files and directories will be owned by you and\n"
            + "# have default permissions.\n"
            + "#\n";

    public final File getDestFile() {
        return destFile;
    }

    public final void setDestFile( final File destFile ) {
        this.destFile = destFile;
    }

    public final boolean isUuencode() {
        return uuencode;
    }

    /**
     * The files to be archived are uuencoded if this is set to true.
     * Defaults to true.
     */
    public final void setUuencode( final boolean uuencode ) {
        this.uuencode = uuencode;
    }

    /**
     * Adds a set of files to be included in the archive.
     */
    public final void addFileset( final FileSet set ) {
        filesets.addElement( set );
    }

    /**
     * Add a script for executing after the unarchival is completed.
     */
    public final void addPostscript( final Postscript script ) {
        postscript = script;
    }

    public final void execute() {
        if ( destFile == null ) {
            throw new BuildException(
                    "destfile must be specified.");
        }
        BufferedOutputStream bos = null;
        BufferedInputStream bis = null;
        try {
            bos = new BufferedOutputStream( new FileOutputStream( destFile ) );
            bos.write( HEADER.getBytes() );
            if ( !filesets.isEmpty() ) {
                for ( int i = 0; i < filesets.size(); i++ ) {
                    final FileSet fs = (FileSet) filesets.elementAt( i );
                    final DirectoryScanner ds = fs.getDirectoryScanner( getProject() );
                    final String[] srcFiles = ds.getIncludedFiles();
                    final File dir = fs.getDir( getProject() );
                    for ( int j = 0; j < srcFiles.length; j++ ) {
                        final String src = srcFiles[j];
                        final String srcWithUnixPath = src.replace( '\\', '/' );
                        final String eofMarker = hashData( srcWithUnixPath.getBytes() );
                        final int idx = srcWithUnixPath.lastIndexOf( '/' );
                        String dirToBeCreated = null;
                        if ( idx >= 0 ) {
                            dirToBeCreated = srcWithUnixPath.substring( 0, idx );
                            bos.write( ( "mkdir -p " + dirToBeCreated + " > /dev/null 2>&1 \n" ).getBytes() );
                        }
                        bos.write( ( "echo x - " + srcWithUnixPath + "\n" ).getBytes() );
                        bos.write( ( "sed >" + srcWithUnixPath + "<< 'END-of-" + eofMarker + "'" ).getBytes() );
                        if ( isUuencode() ) {
                            bos.write( ( ";uudecode -o " + srcWithUnixPath + " " + srcWithUnixPath ).getBytes() );
                        }
                        bos.write( "\n".getBytes() );
                        final File inFile = new File( dir, src );
                        bis = new BufferedInputStream( new FileInputStream( inFile ) );
                        if ( isUuencode() ) {
                            final UUEncoder encoder = new UUEncoder( srcWithUnixPath );
                            encoder.encodeBuffer( bis, bos );
                        } else {
                            copyStream( bis, bos );
                        }
                        bis.close();
                        bos.write( ( "END-of-" + eofMarker + "\n" ).getBytes() );
                    }
                }
            }
            if ( postscript != null ) {
                final String script = getProject().replaceProperties( postscript.txt );
                bos.write( script.getBytes() );
            }
            bos.write( "exit\n".getBytes() );
        } catch ( final IOException ioe ) {
            throw new BuildException( ioe );
        } finally {
            if ( bos != null ) {
                try {
                    bos.close();
                } catch ( final Exception e ) {
                    //ignore
                }
            }
            if ( bis != null ) {
                try {
                    bis.close();
                } catch ( final Exception e ) {
                    //ignore
                }
            }
        }
    }

    private static final void copyStream( final InputStream is,
                                          final OutputStream os ) throws IOException {
        byte[] buffer = new byte[8192];
        int length;
        while ( ( length = is.read( buffer ) ) >= 0 ) {
            os.write( buffer, 0, length );
        }
    }

    private final String hashData( byte[] dataToHash ) {
        return hexStringFromBytes( calculateHash( dataToHash ) );
    }

    private final byte[] calculateHash( byte[] dataToHash ) {
        try {
            final MessageDigest md = MessageDigest.getInstance( "MD5" );
            md.update( dataToHash, 0, dataToHash.length );
            return ( md.digest() );
        } catch ( final NoSuchAlgorithmException e ) {
            return null;
        }
    }

    public final String hexStringFromBytes( byte[] b ) {
        String hex = "";
        int msb;
        int lsb = 0;
        int i;
        for ( i = 0; i < b.length; i++ ) {
            msb = ( (int) b[i] & 0x000000FF ) / 16;
            lsb = ( (int) b[i] & 0x000000FF ) % 16;
            hex = hex + hexChars[msb] + hexChars[lsb];
        }
        return ( hex );
    }

    public static final class Postscript {
        String txt = "";

        public final void addText( final String txt ) {
            this.txt = txt;
        }
    }
}
